/* GStreamer
 * Copyright (C) 2009- Luo Jinghua <sunmoon1997@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more
 */

/**
 * SECTION:element-ppssrc
 *
 * This plugin reads data from a remote location specified by a URI.
 * Supported protocols is 'pps'.
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>
#ifdef HAVE_STDLIB_H
#include <stdlib.h>             /* atoi() */
#endif
#include <string.h>
#include <unistd.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <gst/gstelement.h>
//#include <gst/gst-i18n-plugin.h>
#include <gst/base/gsttypefindhelper.h>
#include "gstscsrc.h"

GST_DEBUG_CATEGORY_STATIC (sopcastsrc_debug);
#define GST_CAT_DEFAULT sopcastsrc_debug

static const GstElementDetails gst_sopcast_src_details =
GST_ELEMENT_DETAILS ("Sopcast stream source",
    "Source/Network",
    "Receive data as a client over the network via sopcast",
    "Luo Jinghua <sunmoon1997@gmail.com>");

static GstStaticPadTemplate srctemplate = GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS_ANY);

enum
{
  PROP_0,
  PROP_LOCATION,
  PROP_USER,
  PROP_PASSWORD
};

static void gst_sopcast_src_uri_handler_init (gpointer g_iface,
    gpointer iface_data);
static void gst_sopcast_src_init (GstSopcastSrc * src,
    GstSopcastSrcClass * g_class);
static void gst_sopcast_src_dispose (GObject * gobject);

static void gst_sopcast_src_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_sopcast_src_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);

static GstFlowReturn gst_sopcast_src_create (GstPushSrc * src, GstBuffer ** buffer);

static gboolean gst_sopcast_src_start (GstBaseSrc * bsrc);

static gboolean gst_sopcast_src_stop (GstBaseSrc * bsrc);

static gboolean gst_sopcast_src_get_size (GstBaseSrc * bsrc, guint64 * size);

static gboolean gst_sopcast_src_is_seekable (GstBaseSrc * bsrc);

static gboolean gst_sopcast_src_do_seek (GstBaseSrc * bsrc,
    GstSegment * segment);
static gboolean gst_sopcast_src_unlock (GstBaseSrc * bsrc);

static gboolean gst_sopcast_src_unlock_stop (GstBaseSrc * bsrc);

static gboolean gst_sopcast_src_set_location (GstSopcastSrc * src,
    const gchar * uri);

static GstStateChangeReturn
gst_sopcast_src_change_state (GstElement * element, GstStateChange transition);

static void
_do_init (GType type)
{
  static const GInterfaceInfo urihandler_info = {
    gst_sopcast_src_uri_handler_init,
    NULL,
    NULL
  };

  g_type_add_interface_static (type, GST_TYPE_URI_HANDLER, &urihandler_info);

  GST_DEBUG_CATEGORY_INIT (sopcastsrc_debug, "sopcastsrc", 0,
      "SCtream src");
}

GST_BOILERPLATE_FULL (GstSopcastSrc, gst_sopcast_src, GstPushSrc,
    GST_TYPE_PUSH_SRC, _do_init);

static void
gst_sopcast_src_base_init (gpointer g_class)
{
  GstElementClass *element_class = GST_ELEMENT_CLASS (g_class);

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&srctemplate));

  gst_element_class_set_details (element_class, &gst_sopcast_src_details);
}

static void
gst_sopcast_src_class_init (GstSopcastSrcClass * klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;
  GstBaseSrcClass *gstbasesrc_class;
  GstPushSrcClass *gstpushsrc_class;

  gobject_class = (GObjectClass *) klass;
  gstbasesrc_class = (GstBaseSrcClass *) klass;
  gstelement_class = (GstElementClass *) klass;
  gstpushsrc_class = (GstPushSrcClass *) klass;

  gobject_class->set_property = gst_sopcast_src_set_property;
  gobject_class->get_property = gst_sopcast_src_get_property;
  gobject_class->dispose = gst_sopcast_src_dispose;

  g_object_class_install_property (gobject_class,
      PROP_LOCATION,
      g_param_spec_string ("location", "Location",
	  "Location to read from", "", G_PARAM_READWRITE));

  gstbasesrc_class->start = GST_DEBUG_FUNCPTR (gst_sopcast_src_start);
  gstbasesrc_class->stop = GST_DEBUG_FUNCPTR (gst_sopcast_src_stop);
  gstbasesrc_class->unlock = GST_DEBUG_FUNCPTR (gst_sopcast_src_unlock);
  gstbasesrc_class->unlock_stop =
      GST_DEBUG_FUNCPTR (gst_sopcast_src_unlock_stop);
  gstbasesrc_class->get_size = GST_DEBUG_FUNCPTR (gst_sopcast_src_get_size);
  gstbasesrc_class->is_seekable =
      GST_DEBUG_FUNCPTR (gst_sopcast_src_is_seekable);
  gstbasesrc_class->do_seek = GST_DEBUG_FUNCPTR (gst_sopcast_src_do_seek);

  gstpushsrc_class->create = GST_DEBUG_FUNCPTR (gst_sopcast_src_create);
  gstelement_class->change_state = GST_DEBUG_FUNCPTR (gst_sopcast_src_change_state);

  GST_DEBUG_CATEGORY_INIT (sopcastsrc_debug, "sopcastsrc", 0,
      "SCtream Source");
}

static void
gst_sopcast_src_init (GstSopcastSrc * src, GstSopcastSrcClass * g_class)
{
  guint major, minor, micro, nano;

  gst_version (&major, &minor, &micro, &nano);

  src->location = NULL;
  src->request_position = 0;
  src->percent = 0;
}

static void
gst_sopcast_src_dispose (GObject * gobject)
{
  GstSopcastSrc *src = GST_SOPCAST_SRC (gobject);

  GST_DEBUG_OBJECT (src, "dispose");

  g_free (src->location);
  src->location = NULL;

  G_OBJECT_CLASS (parent_class)->dispose (gobject);
}

static void
gst_sopcast_src_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  GstSopcastSrc *src = GST_SOPCAST_SRC (object);

  switch (prop_id) {
    case PROP_LOCATION:
    {
      const gchar *location;

      location = g_value_get_string (value);

      if (location == NULL) {
        GST_WARNING ("location property cannot be NULL");
        goto done;
      }
      if (!gst_sopcast_src_set_location (src, location)) {
        GST_WARNING ("badly formatted location");
        goto done;
      }
      break;
    }
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
done:
  return;
}

static void
gst_sopcast_src_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  GstSopcastSrc *src = GST_SOPCAST_SRC (object);

  switch (prop_id) {
    case PROP_LOCATION:
      g_value_set_string (value, src->location);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static gboolean
gst_sopcast_src_connect_server (GstSopcastSrc *src)
{
  struct sockaddr_in sockaddr;
  int sock;

  if (src->socks[0] >= 0) {
    close (src->socks[0]);
    src->socks[0] = -1;
  }

  sock = socket (AF_INET, SOCK_STREAM, 0);
  if (sock < 0) {
    GST_ERROR_OBJECT (src, "Failed to create a socket.");
    return FALSE;
  }

  memset (&sockaddr, 0, sizeof(sockaddr));
  sockaddr.sin_family = AF_INET;
  sockaddr.sin_port = htons (src->ports[1]);
  inet_aton ("127.0.0.1", &sockaddr.sin_addr);

  GST_DEBUG_OBJECT (src, "Connecting to http://127.0.0.1:%d", src->ports[1]);
  if (connect (sock, (const struct sockaddr*)&sockaddr,
	       sizeof(sockaddr)) < 0) {
    close (sock);
    return FALSE;
  }

  src->socks[0] = sock;
  src->first_request = TRUE;
  return TRUE;
}

static gboolean
gst_sopcast_src_send_first_request (GstSopcastSrc *src)
{
  gchar request[1024];
  gchar reply[1024];
  int ret;
  int n;

  snprintf (request, sizeof (request),
	    "GET / HTTP/1.1\r\n"
	    "Host: 127.0.0.1:%d"
	    "User-Agent: gstreamer\r\n"
	    "Accept: */*\r\n"
	    "\r\n", src->ports[1]);

  ret = send (src->socks[0], request, strlen (request), 0);
  if (ret < 0) {
    perror ("Could't sent request to server");
    return FALSE;
  }

  for (n = 0; n < sizeof (reply) - 1; n++) {
    do {
      ret = recv (src->socks[0], &reply[n], 1, 0);
    } while (ret < 0 && errno == EINTR);
    if (ret < 0)
      return FALSE;
    if (n < 4)
      continue;
    if (reply[n] == '\n' && !strncmp (&reply[n - 3], "\r\n\r\n", 4)) {
      reply[n - 3] = '\0';
      GST_DEBUG_OBJECT (src, "Server replied: %s", reply);
      return TRUE;
    }
  }

  GST_ERROR_OBJECT (src, "unexpected return: %s(%d)", reply, n);
  return FALSE;
}

static gchar*
gst_sopcast_src_get_stats (GstSopcastSrc *src)
{
  struct sockaddr_in sockaddr;
  gchar request[] = "state\ns\n";
  gchar reply[2048];
  int ret, n;
  int sock;
  fd_set rfs;
  fd_set efs;
  struct timeval tv;

  sock = socket (AF_INET, SOCK_STREAM, 0);
  if (sock < 0) {
    GST_ERROR_OBJECT (src, "Failed to create a socket.");
    return NULL;
  }

  memset (&sockaddr, 0, sizeof(sockaddr));
  sockaddr.sin_family = AF_INET;
  sockaddr.sin_port = htons (src->ports[1]);
  inet_aton ("127.0.0.1", &sockaddr.sin_addr);

  if (connect (sock, (const struct sockaddr*)&sockaddr,
	       sizeof(sockaddr)) < 0)
    goto failed;

  GST_LOG_OBJECT (src, "Connected to server tcp://127.0.0.1:%d.", src->ports[0]);

  for (n = 0; n < 2; n++) {
    ret = send (sock, request, strlen (request), 0);
    if (ret < 0) {
      perror ("Could't sent request to server");
      goto failed;
    }
    GST_LOG_OBJECT (src, "Sent %d bytes", ret);
  }

  FD_ZERO (&rfs);
  FD_SET (sock, &rfs);
  FD_ZERO (&efs);
  FD_SET (sock, &efs);

  tv.tv_sec = 0;
  tv.tv_usec = 500000;
  ret = select (sock + 1, &rfs, NULL, &efs, &tv);
  if (ret <= 0)
    goto failed;

  for (n = 0; n < sizeof (reply) - 1; n += ret) {
    int i;

    do {
      errno = 0;
      ret = recv (sock, &reply[n], sizeof(reply) - n, 0);
    } while (ret < 0 && errno == EINTR);
    if (ret < 0) {
      perror ("Failed to receive a reply from server");
      break;
    }
    if (ret == 0) {
      GST_LOG_OBJECT (src, "Server don't send a reply");
      break;
    }

    for (i = 0; i < ret; i++) {
      if (reply[n + i] == '\n') {
	close (sock);
	reply[n + i + 1] = '\0';
	GST_LOG_OBJECT (src, "stats: %s", reply);
	return g_strdup (reply);
      }
    }
  }

failed:
  close (sock);
  return NULL;
}

static gboolean
gst_sopcast_src_helper_is_running (GstSopcastSrc *src)
{
  int ret;

  do {
    ret = kill (src->helper_pid, 0);
  } while (ret < 0 && errno == EINTR);
  if (ret == 0)
    return TRUE;

  perror ("helper is not running");
  GST_ERROR_OBJECT (src, "helper is not running.");
  return FALSE;
}

static int
gst_src_src_get_buffering_level (GstSopcastSrc *src)
{
  gchar *stats;
  gint percent = 0;

  stats = gst_sopcast_src_get_stats (src);
  if (stats) {
    gchar **values;

    values = g_strsplit (stats, " ", -1);
    if (values[0] != NULL)
      percent = atoi (values[0]);

    g_strfreev (values);
  }

  g_free (stats);

  return percent;
}

static void
gst_sopcast_src_update_buffering_progress (GstSopcastSrc *src, gint percent)
{
  GstMessage *message;

  if (src->percent == percent)
    return;

  src->percent = percent;

  message = gst_message_new_element (GST_OBJECT_CAST (src),
				     gst_structure_new ("progress", "percent",
							G_TYPE_INT, percent, NULL));
  gst_element_post_message (GST_ELEMENT_CAST (src), message);
}

static GstFlowReturn
gst_sopcast_src_create (GstPushSrc * psrc, GstBuffer ** buffer)
{
  int ret;
  time_t now;
  int percent;
  guint blocksize;
  GstBuffer *buf = NULL;
  GstSopcastSrc *src = GST_SOPCAST_SRC (psrc);

  if (src->first_request) {
    int times = 0;

    src->percent = -1;
    do {
      percent = gst_src_src_get_buffering_level (src);

      if (src->interrupted || !gst_sopcast_src_helper_is_running (src))
	return GST_FLOW_ERROR;
      if (percent > 65)
	break;

      if (percent == src->percent) {
	if (++times > 5)
	  src->percent = -1;
      } else {
	times = 0;
      }

      gst_sopcast_src_update_buffering_progress (src, percent);
      g_usleep (100000);
    } while (1);

    GST_DEBUG_OBJECT (src, "Sending the GET request");
    do {
      if (gst_sopcast_src_connect_server (src))
	break;
      if (src->interrupted || !gst_sopcast_src_helper_is_running (src))
	return GST_FLOW_ERROR;
      g_usleep (500000);
    } while (1);

    GST_DEBUG_OBJECT (src, "Sending the GET request");
    if (!gst_sopcast_src_send_first_request (src)) {
      GST_DEBUG_OBJECT (src, "Couldn't send the GET request");
      goto open_error;
    }
    src->first_request = FALSE;
    src->read_position = 0;
  }

  now = time (NULL);
  if (now - src->last_update > 1 || now - src->last_update < 0) {
    percent = gst_src_src_get_buffering_level (src);
    gst_sopcast_src_update_buffering_progress (src, percent);
    src->last_update = now;
  }

  if (!gst_sopcast_src_helper_is_running (src))
    return GST_FLOW_ERROR;

  blocksize = GST_BASE_SRC (src)->blocksize;
  buf = gst_buffer_try_new_and_alloc (blocksize);
  if (G_UNLIKELY (buf == NULL)) {
    GST_ERROR_OBJECT (src, "Failed to allocate %u bytes", blocksize);
    return GST_FLOW_ERROR;
  }

  do {
    fd_set rfs;
    fd_set efs;
    struct timeval tv;
    int sock = src->socks[0];

    if (src->interrupted /* || !gst_sopcast_src_helper_is_running (src) */)
      goto could_not_read;

    FD_ZERO (&rfs);
    FD_SET (sock, &rfs);
    FD_ZERO (&efs);
    FD_SET (sock, &efs);

    tv.tv_sec = 1;
    tv.tv_usec = 0;
    ret = select (sock + 1, &rfs, NULL, &efs, &tv);
    if (ret < 0)
      continue;

    if (FD_ISSET (sock, &efs))
      goto could_not_read;

    if (!ret || !FD_ISSET (sock, &rfs))
      continue;

    ret = recv (sock, GST_BUFFER_DATA (buf),
		GST_BUFFER_SIZE (buf), 0);

    if (ret < 0 && errno != EINTR)
      goto could_not_read;
    else
      break;
  } while (1);

  GST_LOG_OBJECT (src, "read %d bytes", ret);
  if (ret == 0)
    goto eos;

  if (ret < 0)
    goto could_not_read;

  GST_BUFFER_SIZE (buf) = ret;
  GST_BUFFER_OFFSET (buf) = src->read_position;
  GST_BUFFER_OFFSET_END (buf) = src->read_position + ret;

  *buffer = buf;

  src->read_position += ret;

  return GST_FLOW_OK;
could_not_read:
open_error:
  {
    GST_ELEMENT_ERROR (src, RESOURCE, READ, (NULL), GST_ERROR_SYSTEM);
    if (buf)
      gst_buffer_unref (buf);
    return GST_FLOW_ERROR;
  }
eos:
  {
    GST_DEBUG_OBJECT (src, "hits EOS");
    gst_buffer_unref (buf);
    return GST_FLOW_UNEXPECTED;
  }
}

static int
gst_sopcast_src_allocate_port (void)
{
  int port;
  int sock;
  struct sockaddr_in sockaddr;

  do {
    port = g_random_int_range (3000, 65535);
    sock = socket (AF_INET, SOCK_STREAM, 0);

    memset (&sockaddr, 0, sizeof(sockaddr));
    sockaddr.sin_family = AF_INET;
    sockaddr.sin_port = htons(port);
    inet_aton ("127.0.0.1", &sockaddr.sin_addr);
    if (connect (sock, (const struct sockaddr*)&sockaddr,
		 sizeof(sockaddr)) != 0) {
      close (sock);
      break;
    }
    close (sock);
  } while (1);

  return port;
}

static gboolean
gst_sopcast_src_allocate_ports (GstSopcastSrc * src)
{
  src->ports[0] = gst_sopcast_src_allocate_port ();
  do {
    src->ports[1] = gst_sopcast_src_allocate_port ();
  } while (src->ports[1] == src->ports[0]);

  return src->ports[0] > 0 && src->ports[1] > 0;
}

static gboolean
gst_sopcast_src_spawn_helper (GstSopcastSrc * src)
{
  gchar* argv[12];
  gchar  ports[2][20];
  int argc;
  gboolean ret;

  snprintf (ports[0], sizeof(ports[0]), "%d", src->ports[0]);
  snprintf (ports[1], sizeof(ports[1]), "%d", src->ports[1]);

  argc = 0;
  argv[argc++] = "sp-sc";
  argv[argc++] = src->location;
  argv[argc++] = ports[0];
  argv[argc++] = ports[1];
  argv[argc++] = NULL;

  ret = g_spawn_async ("/tmp", argv, NULL,
		       G_SPAWN_SEARCH_PATH | G_SPAWN_STDOUT_TO_DEV_NULL,
		       NULL, NULL, &src->helper_pid,
		       NULL);
  if (ret)
    return TRUE;

  argv[1] = "sp-sc-auth";
  return g_spawn_async ("/tmp", argv, NULL,
			G_SPAWN_SEARCH_PATH | G_SPAWN_STDOUT_TO_DEV_NULL,
			NULL, NULL, &src->helper_pid,
			NULL);
}

static void
gst_sopcast_src_subproc_main (GstSopcastSrc * src, pid_t ppid, int sock)
{
  fd_set rfs;
  fd_set efs;
  struct timeval tv;
  int i, max, ret;
  char buf[1024];

  if (!gst_sopcast_src_spawn_helper (src))
    _exit (-1);

  max = sysconf(_SC_OPEN_MAX);
  for (i = 3; i < max; i++)
    if (i != sock)
      close (i);

  signal (SIGHUP, SIG_IGN);
  signal (SIGINT, SIG_IGN);

  while (1) {
    int error;

    if (getppid () != ppid) {
       GST_DEBUG_OBJECT (src, "ppid changed %d -> %d", ppid, getppid ());
       break;
    }

    if (!gst_sopcast_src_helper_is_running (src)) {
      break;
    }

    errno = 0;

    FD_ZERO (&rfs);
    FD_SET (sock, &rfs);
    FD_ZERO (&efs);
    FD_SET (sock, &efs);

    tv.tv_sec = 1;
    tv.tv_usec = 0;
    ret = select (sock + 1, &rfs, NULL, &efs, &tv);
    error = errno;

    GST_LOG_OBJECT (src, "select () = %d", ret);
    if (ret == 0)
      continue;
    if (ret < 0 && error == EINTR)
      continue;

    if (FD_ISSET (sock, &efs)) {
       GST_DEBUG_OBJECT (src, "player closed the sock.");
       break;
    }

    if (FD_ISSET (sock, &rfs)) {
      do {
	ret = recv (sock, buf, sizeof (buf), 0);
      } while (ret < 0 && errno == EINTR);
      if (ret < 0 || buf[0] == 'b') {
	GST_DEBUG_OBJECT (src, "player said goodbye to me.");
	break;
      }
    }
  }

  GST_DEBUG_OBJECT (src, "killing sp-sc(%d).", src->helper_pid);
  kill (src->helper_pid, SIGINT);
  kill (src->helper_pid, SIGTERM);
  kill (src->helper_pid, SIGKILL);

  g_spawn_close_pid (src->helper_pid);

  GST_DEBUG_OBJECT (src, "sp-sc is killed.");

  GST_DEBUG_OBJECT (src, "say goodbye back to player.", ppid);
  send (sock, "bye", 3, 0);
  close (sock);

  _exit (0);
}

static gboolean
gst_sopcast_src_spawn_sopcast (GstSopcastSrc * src)
{
  pid_t ppid, pid;
  int ret;
  int fds[2];

  ret = socketpair (AF_UNIX, SOCK_STREAM, 0, fds);
  if (ret < 0)
    return FALSE;

  ppid = getpid ();
  pid = fork ();
  if (pid == 0) {
    close (fds[0]);
    gst_sopcast_src_subproc_main (src, ppid, fds[1]);
  } else if (pid > 0) {
    close (fds[1]);
    src->heatebeat_fd = fds[0];
    src->helper_pid = pid;
  } else {
    close (fds[0]);
    close (fds[1]);
    return FALSE;
  }

  return TRUE;
}

static gboolean
gst_sopcast_src_start (GstBaseSrc * bsrc)
{
  GstSopcastSrc *src;

  src = GST_SOPCAST_SRC (bsrc);

  if (!gst_sopcast_src_allocate_ports (src)) {
    GST_ERROR_OBJECT (src, "Coudn't allocate two ports.");
    return FALSE;
  }
  GST_DEBUG_OBJECT (src, "localport: %d playerport: %d.", src->ports[0], src->ports[1]);

  src->heatebeat_fd = -1;
  if (!gst_sopcast_src_spawn_sopcast (src)) {
    GST_ERROR_OBJECT (src, "Coudn't spawn the sp-sc.");
    return FALSE;
  }

  src->first_request = TRUE;
  src->percent = 0;
  src->last_update = time (NULL);
  return TRUE;
}

static gboolean
gst_sopcast_src_stop (GstBaseSrc * bsrc)
{
  GstSopcastSrc *src;
  int status;
  char buf[12];

  src = GST_SOPCAST_SRC (bsrc);
  GST_DEBUG_OBJECT (src, "stop()");

  close (src->socks[0]);
  src->socks[0] = -1;

  GST_DEBUG_OBJECT (src, "say goodbye to subprocess %d", src->helper_pid);

  send (src->heatebeat_fd, "bye", 3, 0);
  recv (src->heatebeat_fd, buf, 3, 0);
  close (src->heatebeat_fd);
  src->heatebeat_fd = -1;

  GST_DEBUG_OBJECT (src, "killing subprocess %d", src->helper_pid);
  kill (src->helper_pid, SIGINT);
  kill (src->helper_pid, SIGTERM);
  kill (src->helper_pid, SIGKILL);

  waitpid (src->helper_pid, &status, 0);

  src->helper_pid = 0;

  GST_DEBUG_OBJECT (src, "subprocess is killed");

  return TRUE;
}

/* Interrupt a blocking request. */
static gboolean
gst_sopcast_src_unlock (GstBaseSrc * bsrc)
{
  GstSopcastSrc *src;

  src = GST_SOPCAST_SRC (bsrc);
  GST_DEBUG_OBJECT (src, "unlock()");

  src->interrupted = TRUE;
  return TRUE;
}

/* Interrupt interrupt. */
static gboolean
gst_sopcast_src_unlock_stop (GstBaseSrc * bsrc)
{
  GstSopcastSrc *src;

  src = GST_SOPCAST_SRC (bsrc);
  GST_DEBUG_OBJECT (src, "unlock_stop()");

  src->interrupted = FALSE;
  return TRUE;
}

static gboolean
gst_sopcast_src_get_size (GstBaseSrc * bsrc, guint64 * size)
{
  GstSopcastSrc *src;

  src = GST_SOPCAST_SRC (bsrc);

  GST_DEBUG_OBJECT (src, "get_size() = FALSE");
  return FALSE;
}

static gboolean
gst_sopcast_src_is_seekable (GstBaseSrc * bsrc)
{
  GstSopcastSrc *src = GST_SOPCAST_SRC (bsrc);

  GST_DEBUG_OBJECT (src, "seakable: %d", src->seekable);
  return FALSE;
}

static gboolean
gst_sopcast_src_do_seek (GstBaseSrc * bsrc, GstSegment * segment)
{
  GstSopcastSrc *src = GST_SOPCAST_SRC (bsrc);

  GST_DEBUG_OBJECT (src, "do_seek(%" G_GUINT64_FORMAT ")", segment->start);

  if (src->read_position == segment->start)
    return TRUE;

  return FALSE;
}

static GstStateChangeReturn
gst_sopcast_src_change_state (GstElement * element, GstStateChange transition)
{
  GstSopcastSrc *scsrc;
  GstStateChangeReturn result;

  scsrc = GST_SOPCAST_SRC (element);

  switch (transition) {
    case GST_STATE_CHANGE_NULL_TO_READY:
      GST_DEBUG_OBJECT (scsrc, "NULL->READY");
      break;
    case GST_STATE_CHANGE_READY_TO_PAUSED:
       GST_DEBUG_OBJECT (scsrc, "READY->PAUSED");
      break;
    case GST_STATE_CHANGE_PAUSED_TO_PLAYING:
      GST_DEBUG_OBJECT (scsrc, "PAUSED->PLAYING");
      break;
    default:
      break;
  }

  if ((result =
          GST_ELEMENT_CLASS (parent_class)->change_state (element,
              transition)) == GST_STATE_CHANGE_FAILURE)
    goto failure;

  switch (transition) {
    case GST_STATE_CHANGE_PLAYING_TO_PAUSED:
      GST_DEBUG_OBJECT (scsrc, "PLAYING->PAUSED");
      break;
    case GST_STATE_CHANGE_PAUSED_TO_READY:
      GST_DEBUG_OBJECT (scsrc, "PAUSED->READY");
      break;
    case GST_STATE_CHANGE_READY_TO_NULL:
      GST_DEBUG_OBJECT (scsrc, "READY->NULL");
      break;
    default:
      break;
  }

  return result;

  /* ERRORS */
failure:
  {
    GST_DEBUG_OBJECT (scsrc, "parent failed state change");
    return result;
  }
}

static gboolean
gst_sopcast_src_set_location (GstSopcastSrc * src, const gchar * uri)
{
  g_free (src->location);
  src->location = g_strdup (uri);

  return TRUE;
}

static guint
gst_sopcast_src_uri_get_type (void)
{
  return GST_URI_SRC;
}

static gchar **
gst_sopcast_src_uri_get_protocols (void)
{
  static gchar *protocols[] = { "sop", NULL };
  return protocols;
}

static const gchar *
gst_sopcast_src_uri_get_uri (GstURIHandler * handler)
{
  GstSopcastSrc *src = GST_SOPCAST_SRC (handler);

  return src->location;
}

static gboolean
gst_sopcast_src_uri_set_uri (GstURIHandler * handler, const gchar * uri)
{
  GstSopcastSrc *src = GST_SOPCAST_SRC (handler);

  return gst_sopcast_src_set_location (src, uri);
}

static void
gst_sopcast_src_uri_handler_init (gpointer g_iface, gpointer iface_data)
{
  GstURIHandlerInterface *iface = (GstURIHandlerInterface *) g_iface;

  iface->get_type = gst_sopcast_src_uri_get_type;
  iface->get_protocols = gst_sopcast_src_uri_get_protocols;
  iface->get_uri = gst_sopcast_src_uri_get_uri;
  iface->set_uri = gst_sopcast_src_uri_set_uri;
}

static gboolean
plugin_init (GstPlugin * plugin)
{
  return gst_element_register (plugin, "sopcastsrc", GST_RANK_PRIMARY,
      GST_TYPE_SOPCAST_SRC);
}

GST_PLUGIN_DEFINE (GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    "sopcast",
    "sopcast stream src",
    plugin_init, VERSION, "LGPL", PACKAGE, "")
